#include <iostream>
#include <random>
#include <fstream>
#include <sstream>

class Tasks {
  private:
    unsigned int n;
    unsigned int *c;
    unsigned int *d;
    unsigned int *p;
  public:
    Tasks(const char *fname, unsigned int size);
    void print(void);
    bool schedulable(void);
};

Tasks::Tasks(const char *fname, unsigned int size)
{
  std::ifstream in(fname);

  c = new unsigned int[size];
  d = new unsigned int[size];
  p = new unsigned int[size];
  n = 0;
  while(!in.eof() && n < size) {
    char line[80];

    in.getline(line, 80);
    if (!in.eof()) {
      std::stringstream inl(line);

      inl >> c[n] >> d[n] >> p[n];
      if (!inl.fail()) {
        n = n + 1;
      }
    }
  }
}

bool Tasks::schedulable(void)
{
  unsigned int i;

  for (i = 0; i < n; i++) {
    unsigned int j;
    double u, qs;

    u  = 0;
    qs = c[i];
    for (j = 0; j < n; j++) {
      if ((d[j] <= d[i]) && (j != i)) {
        u  += double(c[j]) / double(p[j]);
        qs += double(c[j] * (p[j] - d[j])) / double(p[j]);
      }
    }
    if (u + qs / double(d[i]) > 1) return false;
  }

  return true;
}

void Tasks::print(void)
{
  unsigned int i;

  for (i = 0; i < n; i++) {
    std::cout << c[i] << "," << d[i] << "," << p[i] << std::endl;
  }
}

int main(int argc, char *argv[])
{
  Tasks *ts;

  if (argc < 2) {
    std::cerr << "Usage: " << argv[0] << " <filename>\n";

    return -1;
  }

  ts = new Tasks(argv[1], 16);
  ts->print();
  if (ts->schedulable()) {
    std::cout << "Taskset schedulable!\n";

    return 0;
  } else {
    std::cout << "Taskset NOT schedulable!\n";
  }

  return 1;
}
