#include <iostream>
#include <cstring>
#ifdef DEVEL
#undef NDEBUG
#endif
#include <cassert>

#include "dbg.hh"
#include "simobject.hh"
#include "queuableentity.hh"
#include "priorityqueue.hh"
#include "event.hh"
#include "eventlist.hh"
#include "randomvar.hh"
#include "sporadicevent.hh"
#include "statobject.hh"
#include "stat.hh"
#include "waitobject.hh"
#include "waitstat.hh"
#include "avgobject.hh"
#include "event.hh"
#include "job.hh"
#include "kernel.hh"
#include "task.hh"
#include "cpu.hh"
#include "server.hh"
#include "budget_server.hh"

#include "cbserver.hh"

void CBServer::preempt(long long int t)
{
  if (t > scheduling_deadline && budget > 0) {
    std::cerr << "Missed scheduling deadline for server " << name << " at time " << t << std::endl;
    abort();
  }
  DBG("\tSever descheduled at time " << t << " with dl "
            << scheduling_deadline << std::endl);

  BudgetServer::preempt(t);
}

void CBServer::rechargeBandwidth(long long int t)
{
  DBG("CBS " << name << " postponing at time " << t <<std::endl);

  kernel->block(this, t);

  DBG("\tupdating server deadline from: " << scheduling_deadline
            << " to " << scheduling_deadline + P << std::endl);
  scheduling_deadline += P;
  budget              += Q;

  kernel->unblock(this, t);
}

void CBServer::action(long long int time, Event *e)
{
  if(e == bandwidthExhaust) {
    DBG("CBS " << name << "depleted at time " << time <<std::endl);
    rechargeBandwidth(time);
  }
}

void CBServer::dispatch_task(Task *task, long long int t)
{
  if (current == NULL) {
    if ((scheduling_deadline <= t) || (((long long int)budget * P > Q * (scheduling_deadline - t)))) {
      scheduling_deadline = t + P;
      budget = Q;
      DBG("CBS " << name
                << " received task " << task->name << " at time " << t
                << " - setting server deadline to : "
                << scheduling_deadline << std::endl);
    }
  }
  assert(scheduling_deadline >= t + getBudget(t) || getBudget(t) == 0);
  BudgetServer::dispatch_task(task, t);
}
