#include <cstdint>
#include <cstdlib>
#include <climits>
#ifdef DEVEL
#undef NDEBUG
#endif
#include <cassert>

#include "simobject.hh"
#include "queuableentity.hh"
#include "event.hh"
#include "randomvar.hh"
#include "priorityqueue.hh"
#include "eventlist.hh"
#include "sporadicevent.hh"
#include "statobject.hh"
#include "waitobject.hh"
#include "avgobject.hh"
#include "job.hh"
#include "task.hh"
#include "kernel.hh"
#include "cpu.hh"
#include "prioritykernel.hh"
#include "mqprioritykernel.hh"
#include "gmqprioritykernel.hh"
#include "gmqdelayedprioritykernel.hh"

int GMQDelayedPriorityKernel::pushSchedule(int cpu, long long int t)
{
  int res = 0;

  if ((push_mask == 0) && (pull_mask == 0)) {
    assert(delayed_pushpull == NULL);
    delayed_pushpull = new Event(t + delta);
    delayed_pushpull->setObject(this);
    el->insert(delayed_pushpull);
    res = 1;
  }
  push_mask |= 1 << cpu;

  return res;
}

int GMQDelayedPriorityKernel::pullSchedule(int cpu, long long int t)
{
  int res = 0;

  if ((push_mask == 0) && (pull_mask == 0)) {
    assert(delayed_pushpull == NULL);
    delayed_pushpull = new Event(t + delta);
    delayed_pushpull->setObject(this);
    el->insert(delayed_pushpull);
    res = 1;
  }
  pull_mask |= 1 << cpu;

  return res;
}

void GMQDelayedPriorityKernel::unblock(Task *task, long long int t)
{
  if (task->cpu == -1) task->cpu = 0;
  pushSchedule(task->cpu, t);
  MQPriorityKernel::unblock(task, t);
}

void GMQDelayedPriorityKernel::block(Task *task, long long int t)
{
  pullSchedule(task->cpu, t);
  MQPriorityKernel::block(task, t);
}

void GMQDelayedPriorityKernel::action(long long int time, Event *e)
{
  int i;

  assert(e == delayed_pushpull);

  for (i = 0; i < n_cpus; i++) {
    if (push_mask & (1 << i)) {
      Task *t;

      t = q[i]->getNext();
      if (t) {
        //MQPriorityKernel::block(t, time);
        push(t);
        MQPriorityKernel::unblock(t, time);
      }
    }
  }
  for (i = 0; i < n_cpus; i++) {
    if (pull_mask & (1 << i)) {
      pull(i, time);
    }
  }
  push_mask = 0;
  pull_mask = 0;
  delayed_pushpull = NULL;
}
