#include <cstdlib>
#include <climits>
#ifdef DEVEL
#undef NDEBUG
#endif
#include <cassert>

#include "simobject.hh"
#include "queuableentity.hh"
#include "event.hh"
#include "randomvar.hh"
#include "priorityqueue.hh"
#include "eventlist.hh"
#include "sporadicevent.hh"
#include "statobject.hh"
#include "waitobject.hh"
#include "avgobject.hh"
#include "job.hh"
#include "task.hh"
#include "kernel.hh"
#include "cpu.hh"
#include "prioritykernel.hh"
#include "mqprioritykernel.hh"
#include "gmqprioritykernel.hh"

void GMQPriorityKernel::push(Task *task)
{
  int min = cpus[task->cpu].exec() ? cpus[task->cpu].exec()->getPriority() : INT_MIN;
  int min_cpu = task->cpu;
  int i;

  for (i = 0; i < n_cpus; i++) {
    int p;

    if (cpus[i].exec()) {
      p = cpus[i].exec()->getPriority();
    } else {
      p = INT_MIN;
    }
    if (p < min) {
      min = p;
      min_cpu = i;
    }
  }

  if (min_cpu >= 0) task->cpu = min_cpu;
}

void GMQPriorityKernel::pull(int cpu, long long int t)
{
  Task *max = cpus[cpu].exec();
  int i;

  for (i = 0; i < n_cpus; i++) {
    Task *t1 = q[i]->getNext();

    if ((max == NULL) || (t1 && (t1->getPriority() > max->getPriority()))) {
      if (max && (max->cpu != cpu)) q[max->cpu]->putTask(max);
      max = t1;
    } else {
      if (t1) q[t1->cpu]->putTask(t1);
    }
  }
  if (max && (max->cpu != cpu)) {
    max->cpu = cpu;
    q[max->cpu]->unblock(max, t);
    //MQPriorityKernel::unblock(max, t);
  }
}

void GMQPriorityKernel::unblock(Task *task, long long int t)
{
  if (task->cpu == -1) task->cpu = 0;
  push(task);
  MQPriorityKernel::unblock(task, t);
}

void GMQPriorityKernel::block(Task *task, long long int t)
{
  MQPriorityKernel::block(task, t);
  pull(task->cpu, t);
}
