#include <iostream>
#include <cstring>
#ifdef DEVEL
#undef NDEBUG
#endif
#include <cassert>

#include "dbg.hh"
#include "simobject.hh"
#include "queuableentity.hh"
#include "priorityqueue.hh"
#include "event.hh"
#include "eventlist.hh"
#include "randomvar.hh"
#include "sporadicevent.hh"
#include "statobject.hh"
#include "stat.hh"
#include "waitobject.hh"
#include "waitstat.hh"
#include "avgobject.hh"
#include "event.hh"
#include "job.hh"
#include "kernel.hh"
#include "task.hh"
#include "cpu.hh"
#include "server.hh"
#include "budget_server.hh"
#include "cbserver.hh"
#include "hcbserver.hh"

#include "hcbserver-so.hh"

PriorityQueue HCBServerSO::ss_queue;
HCBServerSO::SuspendedServer *HCBServerSO::ss_first;
long long int HCBServerSO::start_accounting_time       = -1;
long long int HCBServerSO::current_scheduling_deadline = -1;
Event *HCBServerSO::ssExhaust;

void HCBServerSO::ss_account(long long int time)
{
  DBG("\t\t SS account " << ss_first << "\n");
  if (ss_first == NULL) {
    assert(start_accounting_time == -1);
    return;
  }
  if (start_accounting_time == -1) {
    return;
  }
  DBG("\t\t" << ss_first->s->name << " at time " << time);
  DBG("(SAT = " << start_accounting_time << std::endl);
  assert(ss_first->s->scheduling_deadline <= current_scheduling_deadline || current_scheduling_deadline == -1);
  assert(ss_first->s->getBudget(time) >= (time - start_accounting_time));
  assert(ssExhaust != NULL);
  ss_first->s->budget -= (time - start_accounting_time);
  if (ssExhaust) {
    if (eventlist->remove(ssExhaust))
	    delete ssExhaust;
    ssExhaust = NULL;
  }
  DBG("\t\t\t SAT = -1\n");
  start_accounting_time = -1;
}

void HCBServerSO::ss_start_accounting(long long int time)
{
  assert(start_accounting_time == -1);
  assert(ssExhaust == NULL);
  DBG("\t\t SS start accounting: " << ss_first << std::endl);

  if (ss_first == NULL) return;
  DBG("\t\t" << ss_first->s->name << " | " << ss_first->s->scheduling_deadline << " <? " << current_scheduling_deadline << std::endl);
  if (current_scheduling_deadline >= 0 && ss_first->s->scheduling_deadline > current_scheduling_deadline) return;
  
  DBG("\t\t Set at " << ss_first->s->getBudget(time) << std::endl);

  start_accounting_time = time;
  DBG("\t\t\t SAT = " << start_accounting_time << "\n");
  ssExhaust = new Event(time + ss_first->s->getBudget(time));
  ssExhaust->setObject(ss_first->s);
  eventlist->insert(ssExhaust);
  DBG("\t\tshadow exhausted at" << time + ss_first->s->getBudget(time) << std::endl);
}

void HCBServerSO::ss_insert(long long int time)
{
assert(time > 0);
  suspended->setDeadline(scheduling_deadline);
  DBG("\t SS queue insert: " << suspended << std::endl);
  ss_account(time);
  if (getBudget(time) == 0) return;
  if (ss_first != NULL && scheduling_deadline > ss_first->s->scheduling_deadline) {
    if (ss_first) {
      DBG("\t\t" << scheduling_deadline << " > " << ss_first->s->scheduling_deadline);
    } else {
      DBG("\t\t ss first is null");
    }
    DBG(": in queue!" << std::endl);
    ss_queue.insert(suspended);
  } else {
    DBG("\t\tin queue goes " << ss_first << std::endl);
    if (ss_first) ss_queue.insert(ss_first);
    ss_first = suspended;
  }
  ss_start_accounting(time);
}

void HCBServerSO::ss_remove(long long int time)
{
  DBG("\t\tSS Remove " << suspended->s->name << std::endl);
  ss_account(time);
  DBG("\t\tis it " << name << "? (" << ss_first << " " << suspended << ")\n");
  if (ss_first == suspended) {
    DBG("\t\tWas SS FIRST" << std::endl);
    ss_first = CAST(ss_queue.popFirst(), SuspendedServer *);
  } else {
    ss_queue.remove(suspended);
  }
  ss_start_accounting(time);
}

void HCBServerSO::action(long long int time, Event *e)
{
  if (e == bandwidthExhaust) {
    DBG("HCBS-SO " << name << " depleted at time " << time << std::endl);
    kernel->block(this, time);
    replenish = new Event(scheduling_deadline - D + P);
    replenish->setObject(this);
    eventlist->insert(replenish);
  } else if (e == replenish) {
    DBG("HCBS-SO " << name << " replenishing at time " << time << std::endl);
    scheduling_deadline += P;
    budget              += Q;

    kernel->unblock(this, time);
    replenish = NULL;
  } else if (e == ssExhaust) {
    DBG("HCBS-SO" << name << " finished shadow budget at time " << time << std::endl);
    assert(budget == time - start_accounting_time);
    budget = 0;
    ssExhaust = NULL;
    start_accounting_time = -1;
    ss_remove(time);
  } else {
    abort();
  }
}

void HCBServerSO::dispatch_task(Task *task, long long int t)
{
  if (task == NULL) {
    DBG("HCBS-SO" << name << " has noone to serve at time " << t << std::endl);
    assert(current != NULL);
    ss_insert(t);
  } else if (current == NULL) {
    DBG("HCBS-SO" << name << " has new task to serve at time " << t << "(" << budget << ", " << scheduling_deadline << " )" << std::endl);
    ss_remove(t);
    if (t >= scheduling_deadline - D + P) {
      scheduling_deadline = t + D;
      budget = Q;
    } else if (t >= scheduling_deadline) {
      budget = 0;
    }
    DBG("\t\t" << "gave (" << budget << ", " << scheduling_deadline << ")" << std::endl);
  }
  BudgetServer::dispatch_task(task, t);
}

void HCBServerSO::dispatch(long long int t)
{
  ss_account(t);
  current_scheduling_deadline = scheduling_deadline;
  ss_start_accounting(t);
  BudgetServer::dispatch(t);
}

void HCBServerSO::preempt(long long int t)
{
  DBG("HCBServer SO " << name << " preempted at time " << t << std::endl);
  ss_account(t);
  current_scheduling_deadline = -1;
  ss_start_accounting(t);
  CBServer::preempt(t);
}
