#include <iostream>
#include <cstring>
#ifdef DEVEL
#undef NDEBUG
#endif
#include <cassert>

#include "dbg.hh"
#include "simobject.hh"
#include "queuableentity.hh"
#include "priorityqueue.hh"
#include "event.hh"
#include "eventlist.hh"
#include "randomvar.hh"
#include "sporadicevent.hh"
#include "statobject.hh"
#include "stat.hh"
#include "waitobject.hh"
#include "waitstat.hh"
#include "avgobject.hh"
#include "event.hh"
#include "job.hh"
#include "kernel.hh"
#include "task.hh"
#include "cpu.hh"
#include "server.hh"
#include "budget_server.hh"
#include "cbserver.hh"
#include "hcbserver.hh"
#include "accountableentity.hh"

#include "hcbserver2.hh"

void HCBServer2::action(long long int time, Event *e)
{
  if (e == bandwidthExhaust) {
    DBG("HCBS " << name << " depleted at time " << time << std::endl);
    kernel->block(this, time);
    replenish = new Event(scheduling_deadline - D + P);
    replenish->setObject(this);
    eventlist->insert(replenish);
  }

  if (e == replenish) {
    DBG("HCBS " << name << " replenishing at time " << time << std::endl);
    scheduling_deadline += P;
    budget              += Q;

    kernel->unblock(this, time);
  }
}

void HCBServer2::select(HCBServer2 *pi, unsigned int q, long long int d, double t, double &qs, long long int &deadline)
{
  if (pi == this) {
    qs = q;
    deadline = d;
  } else if (pi->current) {
    qs = pi->getBudget(t);
    deadline = pi->scheduling_deadline;
    assert(qs >= 0);
    DBG("\t\t\t\tInterference1: " << pi->name << ": " << qs << ", " << deadline << "| " << (deadline <= t) << "\t");
    if (qs == 0) {
      qs = pi->Q;
      deadline = t + pi->D;
      DBG("-> " << qs << ", " << deadline << "| " << (deadline <= t) << "\t");
    }
  } else {
    qs = pi->Q;
    deadline = pi->D + t;
    DBG("\t\t\t\tInterference2: " << pi->name << ": " << pi->Q << ", " << pi->D + t << "| " << (deadline <= t) << "\t");
  }
  DBG(std::endl);
}

double HCBServer2::dbf1(unsigned int q, long long int d, long long int t, HCBServer2 *pi)
{
  long long int deadline;
  double u, qs;
  HCBServer2 *pj;

  u  = 0;
  select(pi, q, d, t, qs, deadline);
  DBG("\t\t\t" << pi->name << ": " << qs << ", " << deadline << "\n");
  pj = static_cast<HCBServer2 *>(first);
  while(pj) {
    double pj_budget;
    long long int pj_deadline;

    select(pj, q, d, t, pj_budget, pj_deadline);
    DBG(pj_budget << " " << pj_deadline - t << " " << pj->P << std::endl);
    assert(pj_deadline >= t);
    if ((pj_deadline <= deadline) && (pj != pi)) {
      DBG("\t\t\t\t\tu+=" << pj_budget << " / " << double(pj->P) << std::endl);
      DBG("\t\t\t\t\tqs+=" << double(pj_budget) * double(pj->P) << " - " << pj->Q * (pj_deadline - t) <<  " / " << double(pj->P) << std::endl);
      u  += double(pj->Q) / double(pj->P);
      qs += (pj_budget * double(pj->P) - double(pj->Q) * double(pj_deadline - t)) / double(pj->P);
      DBG("\t\t\t\t\t" << u << " " << qs << std::endl);
    }
    pj = static_cast<HCBServer2 *>(pj->nextp);
  }

  return u + qs / double(deadline - t);
}

bool HCBServer2::check(unsigned int q, long long int d, long long int t)
{
  HCBServer2 *pi;

  DBG("\t\tCheck (" << q << ", " << d << ")\n");
  pi = static_cast<HCBServer2 *>(first);
  while(pi) {
    if (dbf1(q, d, t, pi) > 1) return false;
    DBG("\t\t\t" << pi->name << "is OK\n");
    pi = static_cast<HCBServer2 *>(pi->nextp);
  }

  return true;
}

void HCBServer2::dispatch_task(Task *task, long long int t)
{
  if (current == NULL) {
    DBG("DT for " << name << " " << Q << " " << D << " " << P << " at time " << t << std::endl);
    DBG("\t" << name << " " << budget << " " << scheduling_deadline << std::endl);
    total_activations++;
    if (t >= scheduling_deadline - D + P) {
      budget = Q;
      scheduling_deadline = t + D;
      newdl++;
    } else if (t >= scheduling_deadline) {
      budget = 0;
      budget0_1++;
    } else if (!check(budget, scheduling_deadline, t)) {
      if (check(Q, t + D, t)) {
        if (t + D > scheduling_deadline - D + P) {
          budget = Q;
          scheduling_deadline = t + D;
          newdl++;
        } else {
          budget = 0;
          budget0_1++;
        }
      } else {
        if (t > (scheduling_deadline - D + Q)) {
          budget = 0;
        } else {
          budget = Q - (t - (scheduling_deadline - D));
        }
        budget0++;
      }
    } else {
      keep++;
    }
    DBG("\t Gave " << budget << ", " << scheduling_deadline << std::endl);
    DBG("CBS " << name
              << " received task " << task->name << " at time " << t
              << " - setting server deadline to : "
              << scheduling_deadline << std::endl);
    assert(scheduling_deadline - D + P > t);
    last_activation_time = t;
  } else if (task == NULL) {
    assert(last_activation_time >= 0);
  } else abort();
  if (budget && scheduling_deadline < t + getBudget(t)) {
    std::cerr << "Task " << name << " dispatched with d=" << scheduling_deadline << " < " << t << " + " << getBudget(t) << std::endl;
    abort();
  }
  assert(scheduling_deadline >= t + getBudget(t) || budget == 0);

  BudgetServer::dispatch_task(task, t);
}

void HCBServer2::dispatch(long long int t)
{
  DBG("Task " << name << " " << budget << " " << scheduling_deadline<< " dispatched at time " << t << std::endl);
  assert(scheduling_deadline >= t + getBudget(t) || budget == 0);

  BudgetServer::dispatch(t);
}
