#include <iostream>
#include <cstring>
#ifdef DEVEL
#undef NDEBUG
#endif
#include <cassert>

#include "dbg.hh"
#include "simobject.hh"
#include "queuableentity.hh"
#include "priorityqueue.hh"
#include "event.hh"
#include "eventlist.hh"
#include "randomvar.hh"
#include "sporadicevent.hh"
#include "statobject.hh"
#include "stat.hh"
#include "waitobject.hh"
#include "waitstat.hh"
#include "avgobject.hh"
#include "event.hh"
#include "job.hh"
#include "kernel.hh"
#include "task.hh"
#include "cpu.hh"
#include "server.hh"
#include "budget_server.hh"
#include "cbserver.hh"
#include "hcbserver.hh"
#include "accountableentity.hh"
#include "hcbserver2.hh"

#include "hcbserver2r.hh"

double HCBServer2Revised::dbf2(long long int t, HCBServer2Revised *pi, HCBServer2Revised *pk)
{
  long long int deadline;
  double sum, qs;
  HCBServer2Revised *pj;

  select(pi, 0, scheduling_deadline, t, qs, deadline);
  DBG("\t\t\t" << pi->name << ": " << qs << ", " << deadline << "\n");
  sum = qs;
  pj = static_cast<HCBServer2Revised *>(first);
  while(pj) {
    double pj_budget;
    long long int pj_deadline;

    if (pj != pk) {
      select(pj, 0, scheduling_deadline, t, pj_budget, pj_deadline);
      DBG(pj_budget << " " << pj_deadline - t << " " << pj->P << std::endl);
      assert(pj_deadline >= t);
      if ((pj_deadline <= deadline) && (pj != pi)) {
        DBG("\t\t\t\t\tsum +=" << deadline - pj_deadline << " * " << pj->Q << " / " << pj->P << " + " << pj_budget << std::endl);
        sum += (deadline - pj_deadline) * pj->Q / pj->P + pj_budget;
      }
    }
    pj = static_cast<HCBServer2Revised *>(pj->nextp);
  }

  return sum;
}

double HCBServer2Revised::genq(unsigned int q, long long int d, long long int t)
{
  HCBServer2Revised *pi;
  double k, qmin;

  DBG("\t\tGenerating q for " << name << std::endl);
  qmin = -1;
  pi = static_cast<HCBServer2Revised *>(first);
  while(pi) {
    double qs;
    long long int deadline;

    select(pi, q, d, t, qs, deadline);
    if (pi != this && deadline >= scheduling_deadline) {
      double k, q_candidate;

      k = dbf2(t, pi, this);
      DBG("\t\t\t" << pi->name << " has demanded " << k << std::endl);
      DBG("\t\t\t" << (deadline - t) << " - " << k - (double)Q / (double)P * (deadline - scheduling_deadline) << std::endl);
      q_candidate = (deadline - t) - (double)k - (double)Q / (double)P * (deadline - scheduling_deadline);
      DBG("\t\t\t\t" << q_candidate << std::endl);
      if (q_candidate < 0) q_candidate = 0;
      if (qmin < 0 || qmin > q_candidate) {
        qmin = q_candidate;
      }
    }
    DBG("\t\t\tQMin: " << qmin << std::endl);
    pi = static_cast<HCBServer2Revised *>(pi->nextp);
  }
  k = dbf2(t, this, this);
  if (scheduling_deadline - t - k < 0) {
    qmin = 0;
  } else if (qmin < 0 || qmin > scheduling_deadline - t - k) {
    qmin = scheduling_deadline - t - k;
  }
  if (qmin < 0) qmin = 0;

  return qmin;
}

void HCBServer2Revised::dispatch_task(Task *task, long long int t)
{
  if (current == NULL) {
    DBG("DT for " << name << " " << Q << " " << D << " " << P << " at time " << t << std::endl);
    DBG("\t" << name << " " << budget << " " << scheduling_deadline << std::endl);
    total_activations++;
    if (t >= scheduling_deadline - D + P) {
      budget = Q;
      scheduling_deadline = t + D;
      newdl++;
    } else if (t >= scheduling_deadline) {
      budget = 0;
      budget0_1++;
    } else {
      budget = genq(Q, t + D, t);
      if (budget == 0 && check(Q, t + D, t)) {
        budget = Q;
        scheduling_deadline = t + D;
        newdl++;
      }
    }

    DBG("\t Gave " << budget << ", " << scheduling_deadline << std::endl);
    DBG("CBS " << name
              << " received task " << task->name << " at time " << t
              << " - setting server deadline to : "
              << scheduling_deadline << std::endl);
    assert(scheduling_deadline - D + P > t);
    last_activation_time = t;
  } else if (task == NULL) {
    assert(last_activation_time >= 0);
  } else abort();
  if (budget && scheduling_deadline < t + getBudget(t)) {
    std::cerr << "Task " << name << " dispatched with d=" << scheduling_deadline << " < " << t << " + " << getBudget(t) << std::endl;
    abort();
  }
  assert(scheduling_deadline >= t + getBudget(t) || budget == 0);

  BudgetServer::dispatch_task(task, t);
}


