#include <iostream>
#include <cstring>
#ifdef DEVEL
#undef NDEBUG
#endif
#include <cassert>

#include "dbg.hh"
#include "simobject.hh"
#include "queuableentity.hh"
#include "priorityqueue.hh"
#include "event.hh"
#include "eventlist.hh"
#include "randomvar.hh"
#include "sporadicevent.hh"
#include "statobject.hh"
#include "stat.hh"
#include "waitobject.hh"
#include "waitstat.hh"
#include "avgobject.hh"
#include "event.hh"
#include "job.hh"
#include "kernel.hh"
#include "task.hh"
#include "cpu.hh"
#include "server.hh"
#include "budget_server.hh"
#include "cbserver.hh"

#include "hcbserver3.hh"

void HCBServer3::action(long long int time, Event *e)
{
  if (e == bandwidthExhaust) {
    DBG("HCBS " << name << " depleted at time " << time << std::endl);
    kernel->block(this, time);
    replenish = new Event(scheduling_deadline - D + P);
    replenish->setObject(this);
    eventlist->insert(replenish);
  }

  if (e == replenish) {
    DBG("HCBS " << name << " replenishing at time " << time << std::endl);
    scheduling_deadline += P;
    budget              += Q;

    kernel->unblock(this, time);
  }
}

void HCBServer3::dispatch_task(Task *task, long long int t)
{
  if (task == NULL) {
    assert(last_deactivation_time == -1);
    last_deactivation_time = t;
  } else if (current == NULL) {
    DBG("HCHS3: DT for " << name << " " << Q << " " << D << " " << P << " at time " << t << std::endl);
    DBG("\t" << name << " " << budget << " " << scheduling_deadline << std::endl);
    assert(last_deactivation_time > 0 || t == 0);
    assert(t > scheduling_deadline - D);
    if (t >= scheduling_deadline - D + P) {
      budget = Q;
      scheduling_deadline = t + D;
    } else if (t >= scheduling_deadline) {
      budget = 0;
    } else if (budget > t - last_deactivation_time) {
      budget -= (t - last_deactivation_time);
    } else {
      budget = (t - (scheduling_deadline - D)) * Q / P;
      scheduling_deadline = t + D;
    }
    last_deactivation_time = -1;

    DBG("\t Gave " << budget << ", " << scheduling_deadline << std::endl);
    DBG("CBS " << name
              << " received task " << task->name << " at time " << t
              << " - setting server deadline to : "
              << scheduling_deadline << std::endl);
    assert(scheduling_deadline - D + P > t);
  }
  if (budget && scheduling_deadline < t + getBudget(t)) {
    std::cerr << "Task " << name << " dispatched with d=" << scheduling_deadline << " < " << t << " + " << getBudget(t) << std::endl;
    abort();
  }
  assert(scheduling_deadline >= t + getBudget(t) || budget == 0);

  BudgetServer::dispatch_task(task, t);
}

