#include <cstdlib>
#ifdef DEVEL
#undef NDEBUG
#endif
#include <cassert>

#include "simobject.hh"
#include "queuableentity.hh"
#include "event.hh"
#include "randomvar.hh"
#include "priorityqueue.hh"
#include "eventlist.hh"
#include "sporadicevent.hh"
#include "statobject.hh"
#include "waitobject.hh"
#include "avgobject.hh"
#include "job.hh"
#include "task.hh"
#include "kernel.hh"
#include "cpu.hh"
#include "mpprioritykernel.hh"

Task *MPPriorityKernel::schedule(void)
{
  return CAST(ready.popFirst(), Task *);
}

void MPPriorityKernel::unblock(Task *task, long long int t)
{
  int i;

  if ((task->cpu >= 0) && (cpu[task->cpu].exec() == NULL)) {
    cpu[task->cpu].dispatch_task(task, t);
    return;
  }

  ready.insert(task);
  for (i = 0; i < n_cpus; i++) {
    if (cpu[i].exec()) {
      ready.insert(cpu[i].exec());
    }
  }

  for (i = 0; i < n_cpus; i++) {
    cpu[i].prepare_dispatch(schedule(), t);
  }
  for (i = 0; i < n_cpus; i++) {
    cpu[i].finish_dispatch(t);
  }
}

void MPPriorityKernel::block(Task *task, long long int t)
{
  int i;

  for (i = 0; i < n_cpus; i++) {
    if (task == cpu[i].exec()) {
      cpu[i].dispatch_task(schedule(), t);

      return;
    }
  }

  abort();
}

MPPriorityKernel::MPPriorityKernel(CPU *c, int m)
{
  n_cpus = m;
  cpu = c;
}

MPPriorityKernel::~MPPriorityKernel(void)
{
}
