#include <iostream>
#include <cstring>
#ifdef DEVEL
#undef NDEBUG
#endif
#include <cassert>

#include "dbg.hh"
#include "simobject.hh"
#include "queuableentity.hh"
#include "priorityqueue.hh"
#include "event.hh"
#include "eventlist.hh"
#include "randomvar.hh"
#include "sporadicevent.hh"
#include "statobject.hh"
#include "stat.hh"
#include "waitobject.hh"
#include "waitstat.hh"
#include "avgobject.hh"
#include "event.hh"
#include "job.hh"
#include "kernel.hh"
#include "task.hh"
#include "cpu.hh"

#include "server.hh"

int Server::serv_id=0;

Server::Server(Kernel *k, EventList *l) : Task(k, l)
{
  id = serv_id++;

  memset(name, 0, 16);
  memcpy(name, "Serv", 4);
  name[4] = '0' + id / 10;
  name[5] = '0' + id % 10;
  active = false;
}

void Server::action(long long int, Event*)
{
}

void Server::dispatch(long long int t)
{	
  DBG("SERVER " << name << " dispatched at time " << t << std::endl);

  assert(!active);
  assert(current);
  DBG("\tInvoking dispatch on " << current->name << " at time " << t << std::endl);
  active = true;
  current->cpu = cpu;
  current->dispatch(t);
}

void Server::preempt(long long int t)
{	
  DBG("SERVER " << name << " preempted at time " << t << std::endl);

  assert(active);
  if (current) {
    DBG("\ttask " << current->name << " preempt()" << std::endl);
    current->preempt(t);
  }
  active = false;
}

void Server::dispatch_task(Task *task, long long int t)
{
  Task *old_exec = current;

  if (task) {
    DBG("TASK " << task->name << " scheduled on SERVER " << name << " at time "<<t<< std::endl);
  } else {
    DBG("SERVER " << name << " going idle at time "<<t<< std::endl);
  }

  if (task == current) {
    DBG("\tSame task: no ctxt switch!\n");

    return;
  }

  current = task;

  if (!active) {
    DBG("\tServer not active!\n");
    assert(current);
    if (old_exec == NULL) {
      kernel->unblock(this, t);
    }

    return;
  }

  assert(old_exec);
  old_exec->preempt(t);

  if (current) {
    current->cpu = cpu;
    current->dispatch(t);
  } else {
    kernel->block(this, t);
  }
}

Server::~Server(void)
{
}
