#include <cstdlib>
#include <cstring>
#ifdef DEVEL
#undef NDEBUG
#endif
#include <cassert>

#include "simobject.hh"
#include "queuableentity.hh"
#include "priorityqueue.hh"
#include "event.hh"
#include "eventlist.hh"
#include "randomvar.hh"
#include "sporadicevent.hh"
#include "statobject.hh"
#include "stat.hh"
#include "waitobject.hh"
#include "waitstat.hh"
#include "avgobject.hh"
#include "event.hh"
#include "job.hh"
#include "kernel.hh"
#include "task.hh"

int Task::task_id;

void Task::action(long long int time, Event *e)
{
  if (e == arrival) {
    Job *j = new Job(time, exec->sample(), time + relative_deadline, response_time);

    trace(arrival, this, time);
//std::cerr << "Wake " << task->name << " up at time " << t << std::endl;
    if (current_job) {
      pending_jobs.insert(j);
    } else {
      current_job = j;
      kernel->unblock(this, time);
    }
  } else if (current_job && e == current_job->end) {
//std::cerr << "J " << current_job->r << " " << time << " (" << current_job->d << ")" << std::endl;
    trace(finish, this, time);
    current_job->finish(time);
    kernel->block(this, time);
    delete current_job;
    if (migrations) migrations->reset();
    current_job = CAST(pending_jobs.popFirst(), Job *);
    if (current_job) {
      kernel->unblock(this, time);
    }
  } else {
    abort();
  }
}

void Task::dispatch(long long int t)
{
//if (exec[i]) std::cerr << "Schedule " << exec[i]->name << " on CPU " << i << std::endl;
//std::cerr << "Schedule " << name << " on CPU " << cpu << std::endl;
//if (old_cpu != cpu) std::cerr << "Migrate task " << name << " from " << old_cpu << " to " << cpu << " at time " << t << std::endl;
  assert(dispatch_time == -1);
  assert(current_job->end == NULL);
  trace(dispatch, this, t);
  if (migrations && (old_cpu != cpu)) {
    migrations->inc();
  }
  old_cpu = cpu;
  dispatch_time = t;
  current_job->end = new Event(t + current_job->c);
  current_job->end->setObject(this);	// FIXME?
  eventlist->insert(current_job->end);
}

void Task::preempt(long long int t)
{
  assert(dispatch_time >= 0);
  assert(current_job->c >= (t - dispatch_time));
  trace(preempt, this, t);
  if (eventlist->remove(current_job->end)) {
    delete current_job->end;
  }
  current_job->end = NULL;
  current_job->c -= (t - dispatch_time);
  dispatch_time = -1;
}

Task::Task(Kernel *k, EventList *l)
{
  kernel = k;
  eventlist = l;
  priority = -1;
  dispatch_time = -1;
  old_cpu = -1;
  cpu = -1;
  current_job = NULL;
}

Task::Task(RandomVar *c, RandomVar *interarrival, int dl, Kernel *k, EventList *l)
{
  relative_deadline = dl;
  kernel = k;
  exec = c;
  if (interarrival) {
    arrival = new SporadicEvent(0, interarrival, l);
    arrival->setObject(this);
  }
  eventlist = l;
  priority = -1;
  response_time = NULL;
  migrations = NULL;
  current_job = NULL;
  dispatch_time = -1;
  id = task_id++;
  old_cpu = -1;
  cpu = -1;
  memset(name, 0, 16);
  memcpy(name, "Task", 4);
  name[4] = '0' + id / 10;
  name[5] = '0' + id % 10;
}

Task::~Task(void)
{
  Job *j;

  delete current_job;
  while ((j = (Job *)pending_jobs.popFirst())) {
    delete j;
  }
}

