#include <iostream>
#include <sstream>
#include <fstream>
#include <cstring>

#include "tracer.hh"
#include "randomvar.hh"
#include "filevar.hh"
#include "simobject.hh"
#include "queuableentity.hh"
#include "priorityqueue.hh"
#include "event.hh"
#include "eventlist.hh"
#include "sporadicevent.hh"
#include "statobject.hh"
#include "stat.hh"
#include "waitobject.hh"
#include "waitstat.hh"
#include "distribution.hh"
#include "waitdistribution.hh"
#include "avgobject.hh"
#include "avgstat.hh"
#include "job.hh"
#include "kernel.hh"
#include "task.hh"
#include "tracedtask.hh"
#include "taskset.hh"

static unsigned long long int gcd(unsigned long long int a, unsigned long long int b)
{
  while(1) {
/*
    if (a == 0) return b;
    b %= a;
    if (b == 0) return a;
    a %= b;
*/
    if (a > b) {
      if (b == 0) return a;
      a = a % b;
    } else {
      if (a == 0) return b;
      b = b % a; 
    }
  }
}

unsigned long long int TaskSet::lcm(unsigned long long int a, unsigned long long int b)
{
  return (a / gcd(a, b)) * b;
}

bool TaskSet::deterministic_task(const char *l, unsigned int *c_i, unsigned int *d_i, unsigned int *p_i)
{
  std::stringstream in(l);

  in >> *c_i >> *d_i >> *p_i;

  if (in.fail()) {
    in.str(l);
    in.clear();
    in >> *c_i >> *p_i;
    *d_i = *p_i;
  }

  if (*c_i == 0) *c_i = 1;

  return !in.fail();
}

bool TaskSet::probabilistic_task(const char *l, char *fname, unsigned int *d_i, unsigned int *p_i)
{
  std::stringstream in(l);

  in >> fname >> *d_i >> *p_i;

  if (in.fail()) {
    in.str(l);
    in.clear();
    in >> fname >> *p_i;
    *d_i = *p_i;
  }

  return !in.fail();
}

TaskSet::TaskSet(const char *fname)
{
  std::ifstream in(fname);
  size = 0;
  trc = NULL;
  rstat = NULL;
  mstat = NULL;

  if (!in.is_open()) {
    c = NULL;
    p = NULL;
    d = NULL;
    prio = NULL;
    tasks = NULL;

    return;
  }

  c = new RandomVar*[max_tasks];
  p = new RandomVar*[max_tasks];
  d = new unsigned int[max_tasks];
  prio = new int[max_tasks];
  hyper_p = 1;
  while(!in.eof()) {
    unsigned int c_i, d_i, p_i;
    char cdf_name[16];
    char line[80];

    in.getline(line, 80);
    if (!in.eof()) {
      if (deterministic_task(line, &c_i, &d_i, &p_i)) {
        c[size] = new Delta(c_i);
        p[size] = new Delta(p_i);
        d[size] = d_i;
        prio[size] = size;	// Assume tasks are already ordered in the file...
        size++;
      } else if (probabilistic_task(line, cdf_name, &d_i, &p_i)) {
        c[size] = new FileVar(cdf_name);
        p[size] = new Delta(p_i);
        d[size] = d_i;
        prio[size] = size;	// Assume tasks are already ordered in the file...
        size++;
      } else {
        std::cout << "Error reading taskset!!!\n";
        delete[] c;
        delete[] p;
        delete[] prio;
        c = NULL;
        p = NULL;
        prio = NULL;
        tasks = NULL;

        return;
      }
      hyper_p = lcm(hyper_p, p_i);
    }
  }
  tasks = new Task*[size];
  memset(tasks, 0, size * sizeof(Task *));
}

TaskSet::~TaskSet()
{
  unsigned int i;

  for (i = 0; i < size; i++) {
    delete c[i];
    delete p[i];
    if (mstat) {
      if (mtype == stat_average) { 
        delete (AvgStat *)mstat[i];
      } else if (rtype == stat_distribution) {
        //delete (AvgDistribution *)mstat[i];
      }
    }
    if (rstat) {
      if (rtype == stat_average) { 
        delete (WaitStat *)rstat[i];
      } else if (rtype == stat_distribution || rtype == stat_dlmiss) {
        delete (WaitDistribution *)rstat[i];
      }
    }
    if (tasks[i]) delete tasks[i];
  }
  delete[] c;
  delete[] p;
  delete[] d;
  delete[] tasks;
  delete[] prio;
  delete[] mstat;
  delete[] rstat;
  std::cout << "Closing taskset of size " << size << std::endl;
}

unsigned long long int TaskSet::getHyperPeriod(void)
{
  return hyper_p;
}

unsigned int TaskSet::getSize(void)
{
  return size;
}

int TaskSet::addToKernel(Kernel *k, EventList *l)
{
  unsigned int i;

  for (i = 0; i < size; i++) {
    if (trc) {
      tasks[i] = new TracedTask(c[i], p[i], d[i], k, l, trc);
    } else {
      tasks[i] = new Task(c[i], p[i], d[i], k, l);
    }
    tasks[i]->setPriority(prio[i]);
    if (rstat) tasks[i]->setR(rstat[i]);
    if (mstat) tasks[i]->setM(mstat[i]);
  }

  return size;
}

WaitObject **TaskSet::getRStat(StatType type, unsigned int scale)
{
  unsigned int i;
  rstat = new WaitObject*[size];

  rtype = type;
  for (i = 0; i < size; i++) {
    if (type == stat_average) {
      rstat[i] = new WaitStat;
    } else if (type == stat_distribution) {
      rstat[i] = new WaitDistribution(scale);
    } else if (type == stat_dlmiss) {
      rstat[i] = new WaitDistribution(d[i]);
    } else rstat[i] = NULL;
  }

  return rstat;
}

AvgObject **TaskSet::getMStat(StatType type)
{
  unsigned int i;
  mstat = new AvgObject*[size];

  mtype = type;
  for (i = 0; i < size; i++) {
    if (type == stat_average) {
      mstat[i] = new AvgStat;
    } else if (type == stat_distribution) {
      //mstat[i] = new AvgDistribution;
      mstat[i] = NULL;
    } else mstat[i] = NULL;
  }

  return mstat;
}

void TaskSet::print(std::ostream &o)
{
  unsigned int i;

  for (i = 0; i < size; i++) {
    char cname[32], pname[32];
    unsigned int pint;

    memset(cname, 0, 32);
    memset(pname, 0, 32);
    c[i]->dump(cname);
    p[i]->dump(pname);
    pint = atoi(pname);
    if (d[i] == pint) {
      o << i << " - (" << cname << ", " << pname << ")\n";
    } else {
      o << i << " - (" << cname << ", " << d[i] << ", " << pname << ")\n";
    }
  }
}
