#include <sstream>
#include <random>

#include "randomvar.hh"
#include "poissonvar.hh"

PoissonVar::PoissonVar(double mn)
{
  mean = mn;
}

long long int PoissonVar::sample(void)
{
  static const double EULER = 2.71828182845904523536;
  double r = ((double) rand() / (RAND_MAX));
  double cumulative = 0;
  int n = 0;

  while(cumulative < r) {
    cumulative += pow(EULER, -mean) * (pow(mean, n) / fact(n));
    n++;
  }
	
  return n;
}

long long int PoissonVar::fact(unsigned int n)
{
  long long int res = 1;
  unsigned int i;

  for(i = 2; i <= n; i++) {
    res = i * res;
  }
	
  return res;
}

void PoissonVar::dump(char *buff)
{
  std::stringstream out;

  out.rdbuf()->pubsetbuf(buff, 32/*FIXME!*/);
  out << "PoissonVar";
}
