#include <cstdlib>
#ifdef DEVEL
#undef NDEBUG
#endif
#include <cassert>

#include "queuableentity.hh"
#include "priorityqueue.hh"

void PriorityQueue::ConsistencyCheck(void)
{

#ifdef DEVEL
  QueuableEntity *temp;

  temp = first;
  while(temp) {
    assert(temp->next != temp);
    temp = temp->next;
  }
#endif
}

#ifdef DEVEL
bool PriorityQueue::contains(QueuableEntity * e)
{
  QueuableEntity *temp;

  ConsistencyCheck();
  temp = first;
  while (temp != NULL) {
    if (temp == e) return true;
    temp = temp->next;
  }
  ConsistencyCheck();

  return false;
}
#endif

PriorityQueue::PriorityQueue(void)
{
  first = NULL;
  size = 0;
}

QueuableEntity* PriorityQueue::popFirst(void)
{
  QueuableEntity* res = first;

  ConsistencyCheck();
  if (res) {
    first = res->next;
    size--;
  }
  ConsistencyCheck();

  return res;
}

int PriorityQueue::remove(const QueuableEntity * en)
{
  QueuableEntity *temp, *prev;                 

  ConsistencyCheck();
  temp = first;
  prev = NULL;
  while(temp && (temp != en)) {
    prev = temp;
    temp = temp->next;
  }
  ConsistencyCheck();
  if (temp == NULL) {
    return 0;
  }

  size--;
  if (temp == first) {
    first = temp->next;
  } else {
    prev->next = temp->next;
  }
  ConsistencyCheck();

  return 1;
}

void PriorityQueue::insert(QueuableEntity * en)
{
  QueuableEntity *temp, *prev;                 

  assert(!contains(en));
  ConsistencyCheck();
  temp = first;
  prev = NULL;
  while(temp && (*temp < *en)) {
    prev = temp;
    temp = temp->next;
  }
  ConsistencyCheck();

  if(temp == first) {
    en->next = first;
    first = en;
  } else {
    prev->next = en;
    en->next = temp;
  }
  size++;
  ConsistencyCheck();
}

void PriorityQueue::process(bool (*f)(QueuableEntity *, void *), void *parm)
{
  QueuableEntity *p;

  ConsistencyCheck();
  for (p = first; p; p = p->next) {
    if (!f(p, parm)) break;
  }
  ConsistencyCheck();
}

#if 0
void PriorityQueue::remove(QueuableEntity * e)
{
    if (contains(e)){
        QueuableEntity *temp;

        if (e->getId() == element->getId()) {
            temp = element;
            element=element->next();
   //delete temp;
        } else { 
            int i=0;

            temp=element;
            while (i < size - 1) {
                if (temp->next()->getId() == e->getId()) break;
                temp=temp->next();
                i++;
            }
            if (temp == NULL) return;
            temp->next = temp->next->next();
  //delete e;
        }
        size--;
    }
}

void PriorityQueue::print(void)
{
  QueuableEntity *curr = first;
  int i = 0;

  std::cout << "LIST " << std::endl;
  while (curr!=NULL){
    std::cout << curr->getPriority() << std::endl;
    curr=curr->next;
    i++;
  }
  std::cout << std::endl;
}

#endif

