#include <iostream>
#include <fstream>
#include <cstdbool>
#include <cstring>
#include <unistd.h>

#include "simobject.hh"
#include "queuableentity.hh"
#include "event.hh"
#include "priorityqueue.hh"
#include "eventlist.hh"
#include "periodicevent.hh"
#include "randomvar.hh"
#include "uniformvar.hh"
#include "sporadicevent.hh"
#include "statobject.hh"
#include "stat.hh"
#include "distribution.hh"
#include "waitobject.hh"
#include "waitstat.hh"
#include "waitdistribution.hh"
#include "avgobject.hh"
#include "avgstat.hh"
#include "job.hh"
#include "task.hh"
#include "task2.hh"
#include "tracer.hh"
#include "txttracer.hh"
#include "tracedtask.hh"
#include "cpu.hh"
#include "kernel.hh"
#include "taskset.hh"
#include "prioritykernel.hh"
#include "mpprioritykernel.hh"
#include "mqprioritykernel.hh"
#include "gmqprioritykernel.hh"
#include "gmqdelayedprioritykernel.hh"
#include "edfkernel.hh"
#include "edfmpkernel.hh"
#include "edfmqkernel.hh"
#include "edfgmqkernel.hh"
#include "edfgmqdelayedkernel.hh"
#include "server.hh"
#include "budget_server.hh"
#include "cbserver.hh"
#include "hcbserver.hh"
#include "accountableentity.hh"
#include "hcbserver2.hh"
#include "hcbserver2r.hh"
#include "hcbserver3.hh"
#include "hcbserver-so.hh"

static double cr, alpha = 0.9, sz = 1;
static int server_type = 1;
static unsigned long long int reporting_period;
static bool multiqueue;
static bool edf;
static unsigned int cpus = 1;
static Tracer *trc;
static unsigned int n_of_hyperperiods = 2;
static unsigned long long int finishing_time;
static int stats;
static int pushpull_delay = -1;

class ServerSet : public TaskSet {
    RandomVar **b;
    Server   **cbs;
    Kernel   **kern;
    unsigned int *qs;
    unsigned int *ts;
    unsigned int type;
    unsigned int *task_type;
  public:
    ServerSet(const char *fname, unsigned int type, double cr = 0, double alpha = 0, double s = 0);
    ~ServerSet();
    int addToKernel(Kernel *k, EventList *l);
    WaitObject **getRStat(StatType type, unsigned int scale = 1);
    void print(std::ostream &o);
};

ServerSet::ServerSet(const char *fname, unsigned int t, double cr, double alpha, double s)
{
  std::ifstream in(fname);
  size = 0;
  trc = NULL;
  rstat = NULL;
  mstat = NULL;
  tasks = NULL;
  cbs = NULL;
  kern = NULL;

  if (!in.is_open()) {
    c = NULL;
    p = NULL;
    d = NULL;
    prio = NULL;
    qs = NULL;
    ts = NULL;

    return;
  }

  c = new RandomVar*[max_tasks];
  p = new RandomVar*[max_tasks];
  b = new RandomVar*[max_tasks];
  d = new unsigned int[max_tasks];
  prio = new int[max_tasks];
  hyper_p = 1;
  qs = new unsigned int[max_tasks];
  ts = new unsigned int[max_tasks];
  task_type = new unsigned int[max_tasks];
  while(!in.eof()) {
    unsigned int c_i, d_i, p_i;
    char line[80];

    in.getline(line, 80);
    if (!in.eof()) {
      if (deterministic_task(line, &c_i, &d_i, &p_i)) {
        int csize, psize, cmean, pmean, cmin, pmin;

        task_type[size] = size % 3;
        if (cr) {
          cmean = cr * c_i;
          pmean = (long long int)cmean * (long long int)p_i / (alpha * c_i);
          csize = c_i * s;
          psize = p_i * s;
        } else {
          cmean = 5 * c_i / 6;
          csize = c_i;
          pmean = 7 * p_i / 6;
          psize = p_i;
        }
        if (pmean < psize / 2 + 1) {
          pmean = psize / 2 + 1;
	}
        if (cmean < csize / 2 + 1) {
          cmean = csize / 2 + 1;
	}
	pmin = pmean - psize / 2;
        cmin = cmean - csize / 2;
        switch (task_type[size]) {
          case 2:
            c[size] = new Delta(c_i);
            p[size] = new Delta(p_i);
            break;
          case 1:
            c[size] = new UniformVar(c_i - (c_i / 3), c_i);
            p[size] = new UniformVar(p_i, p_i + (p_i / 3));
            break;
          case 0:
            c[size] = new UniformVar(cmin, cmean + csize / 2);
            p[size] = new UniformVar(pmin, pmean + psize / 2);
            //c[size] = new UniformVar(c_i - 2 * (c_i / 3), c_i + c_i / 3);
            //p[size] = new UniformVar(p_i - p_i / 3, p_i + 2 * (p_i / 3));
            break;
          default:
            std::cerr << "Uh??? Type = " << task_type[size] << std::endl;
            exit(-1);
        }
        b[size] = new UniformVar(c_i / 10, 2 * c_i / 10);
#if 0 
        std::cout << "T" << size << ": (" << c_i - 2 * (c_i / 3) << " - " << c_i + c_i / 3
                         << ", " << p_i - p_i / 3 << " - " << p_i + 2 * (p_i / 3)
                         << ") --- (" << c_i << ", " << d_i << ", " << p_i << ")\n";
#endif
        d[size] = d_i;
        qs[size] = c_i;
        ts[size] = p_i;
        prio[size] = size;	// Assume tasks are already ordered in the file...
        size++;
      } else {
        std::cout << "Error reading taskset!!!\n";
        delete[] c;
        delete[] p;
        delete[] prio;
        delete[] qs;
        delete[] ts;
        delete[] task_type;
        c = NULL;
        p = NULL;
        prio = NULL;
        qs = NULL;
        ts = NULL;

        return;
      }
      hyper_p = lcm(hyper_p, p_i);
    }
  }
  tasks = new Task*[size];
  memset(tasks, 0, size * sizeof(Task *));
  cbs = new Server *[size];
  memset(cbs, 0, size * sizeof(Server *));
  kern = new Kernel *[size];
  memset(kern, 0, size * sizeof(Kernel *));
  type = t;
}

int ServerSet::addToKernel(Kernel *k, EventList *l)
{
  unsigned int i;

  for (i = 0; i < size; i++) {
    if (type == 1) {
      cbs[i] = new HCBServer(qs[i], ts[i], k, l);
    } else if (type == 2) {
      cbs[i] = new HCBServer2(qs[i], d[i], ts[i], k, l);
    } else if (type == 3) {
      cbs[i] = new HCBServerSO(qs[i], d[i], ts[i], k, l);
    } else if (type == 4) {
      cbs[i] = new HCBServer3(qs[i], d[i], ts[i], k, l);
    } else if (type == 5) {
      cbs[i] = new HCBServer2Revised(qs[i], d[i], ts[i], k, l);
    } else {
      abort();
    }
    kern[i] = new PriorityKernel(cbs[i]);
    if (trc) {
      tasks[i] = new TracedTask(c[i], p[i], d[i], kern[i], l, trc);
    } else {
      tasks[i] = new Task(c[i], p[i], d[i], kern[i], l);
      //tasks[i] = new Task2(c[i], p[i], b[i], d[i], kern[i], l);
    }

    if (rstat) tasks[i]->setR(rstat[i]);
    if (mstat) tasks[i]->setM(mstat[i]);
  }

  return size;
}

void ServerSet::print(std::ostream &o)
{
  unsigned int i;

  for (i = 0; i < size; i++) {
    o << i << " - (" << qs[i] << ", " << d[i] << ", " << ts[i] << ") | " << task_type[i] << "\n";
  }
  TaskSet::print(o);
}

WaitObject **ServerSet::getRStat(StatType type, unsigned int scale)
{
  if (type == stat_dlmiss) {
    unsigned int i;

    rstat = new WaitObject*[size];
    rtype = type;
    for (i = 0; i < size; i++) {
      rstat[i] = new WaitDistribution(ts[i], d[i]);
    }

    return rstat;
  } else return TaskSet::getRStat(type, scale);
}


ServerSet::~ServerSet()
{
  unsigned int i;

  for (i = 0; i < size; i++) {
    delete cbs[i];
    delete kern[i];
  }
  delete[] cbs;
  delete[] kern;
  delete[] qs;
  delete[] ts;
  delete[] task_type;
}

class T: public SimObject {
    long long int end;
  public:
    T(long long int e) {end = e;};
    virtual ~T(void) {};
    virtual void action(long long int time, Event *e);
};

void T::action(long long int time, Event *)
{
  std::cerr << "T = " << time << " (" << double(time) / double(end) << ")" << std::endl;
}

static const char *tmp_name;
static double tmp_sum;
static unsigned int tmp_samples;
static bool dist_print(long long int v, double p)
{
  //std::cout << "P(" << v << ") = " << p << std::endl;
  tmp_sum += p;
  std::cout << tmp_name << "\t" << v << " " << p << " " << tmp_sum;
  std::cout << " " << p * tmp_samples << " " << tmp_sum * tmp_samples << std::endl;

  return true;
}

static void print_results(StatObject *s, const char *n, const char *sname)
{
  Stat *stat;
  Distribution *distro;

#ifdef __GXX_RTTI
  stat   = dynamic_cast<Stat *>(s);
  distro = dynamic_cast<Distribution *>(s);
#else
  stat   = reinterpret_cast<Stat *>(s);
  distro = NULL;
#endif

  if (stat) {
    std::cout << "[" << n << "] Number of events: "        << stat->getNum()          << std::endl;
    std::cout << "[" << n << "] Min " << sname << ": "           << stat->getMin()          << std::endl;
    std::cout << "[" << n << "] Max " << sname << ": "           << stat->getMax()          << std::endl;
    std::cout << "[" << n << "] Avg " << sname << ": "           << stat->getAverage()      << std::endl;
    std::cout << "[" << n << "] Variance: "                << stat->getVariance()     << std::endl;
    std::cout << "[" << n << "] 90% confidence interval: " << stat->getConfidence(90) << std::endl;
    std::cout << "[" << n << "] 95% confidence interval: " << stat->getConfidence(95) << std::endl;
  }

  if (distro) {
    std::cout << "Stream " << n << " " << distro->getN() << std::endl;
    tmp_name = n;
    tmp_sum = 0;
    tmp_samples = distro->getN();
    distro->process(dist_print);
  }
}
    
static void print_help(const char *name)
{
  std::cerr << "Usage: " << name << " <taskset> [<num CPUs>]\n";
}

static const char *args_parse(int argc, char *argv[])
{
  int opt;

  while ((opt = getopt(argc, argv, "Rrsf:l:t:n:p:d:me2345C:A:S:")) != -1) {
    switch (opt) {
      case 'C':
        cr = atof(optarg);
        break;
      case 'A':
        alpha = atof(optarg);
        break;
      case 'S':
        sz = atof(optarg);
        break;
      case 'n':
        cpus = atoi(optarg);
        break;
      case 'p':
        reporting_period = atoll(optarg);
        break;
      case 'm':
        multiqueue = true;
        break;
      case 'e':
        edf = true;
        break;
      case 't':
        trc = new TxtTracer(optarg);
        break;
      case 'l':
        n_of_hyperperiods = atoi(optarg);
        break;
      case 'f':
        finishing_time = atoll(optarg);
        break;
      case 'd':
        pushpull_delay = atoi(optarg);
        break;
      case 'r':
        stats |= 1;
        break;
      case 'R':
        stats |= 4;
        break;
      case 's':
        stats |= 2;
        break;
      case '2':
        server_type = 2;
        break;
      case '3':
        server_type = 3;
        break;
      case '4':
        server_type = 4;
        break;
      case '5':
        server_type = 5;
        break;
      default:
        print_help(argv[0]);
        exit(EXIT_FAILURE);
    }
  }

  if (optind >= argc) {
    print_help(argv[0]);

    exit(EXIT_FAILURE);
  }

  return argv[optind];
}

int main(int argc, char *argv[])
{
  long long int hyperperiod;
  EventList l;
  Kernel *kern;
  ServerSet *ts;
  T *o;
  PeriodicEvent *s = NULL;
  WaitObject **rstat = NULL;
  AvgObject **mstat = NULL;
  const char *fname;
  CPU *c;
  WaitStat idle;

  fname = args_parse(argc, argv);
  ts = new ServerSet(fname, server_type, cr, alpha, sz);
  ts->print(std::cout);

  if (trc) {
    ts->setTracer(trc);
  }
  if (stats & 1) {
    rstat = ts->getRStat(TaskSet::stat_distribution, 100);
  } else if (stats & 4) {
    rstat = ts->getRStat(TaskSet::stat_dlmiss);
  }
  if (stats & 2) {
    mstat = ts->getMStat(TaskSet::stat_average);
  }

  c = new CPU(&idle);
  kern = new EDFKernel(c);
  ts->addToKernel(kern, &l);

  hyperperiod = ts->getHyperPeriod();

  if (finishing_time == 0) {
    finishing_time = hyperperiod * n_of_hyperperiods;
    std::cout << "Simulating " << n_of_hyperperiods << " HyperPeriods: " << finishing_time << std::endl;
  }
  if (reporting_period) {
    o = new T(finishing_time);
    s = new PeriodicEvent(0, reporting_period, &l);
    s->setObject(o);
  } else {
    o = NULL;
  }
  l.run(finishing_time);

  if (rstat) {
    unsigned int i;
    char name[16];

    memset(name, 0, 16);
    memcpy(name, "Task", 4);

    for (i = 0; i < ts->getSize(); i++) {
      name[4] = '0' + i / 10;
      name[5] = '0' + i % 10;
      print_results((StatObject *)rstat[i], name, "response time"); //FIXME: Name
    }
  }
  if (mstat) {
    unsigned int i;
    char name[16];

    memset(name, 0, 16);
    memcpy(name, "Task", 4);

    for (i = 0; i < ts->getSize(); i++) {
      name[4] = '0' + i / 10;
      name[5] = '0' + i % 10;
      print_results((StatObject *)mstat[i], name, "migrations"); //FIXME: Name
    }
  }
  std::cout.precision(20);
  std::cout << "CPU idle time: " << idle.getAverage() * idle.getNum() << std::endl;

  delete ts;
  delete kern;
  delete c;
  delete o;

  return 0;
}
